xquery version "3.1";

(:~
 : This module is responsible for pulling the whole transformation/serialization
 : of the edited text together.
 :
 : It creates all the single HTML files for each notebook as well as a complete
 : intermediate format TEI/XML that is used for the print output.
 :
 : @author Michelle Weidling
 : @version 1.0
 :)

module namespace etTransfo="http://fontane-nb.dariah.eu/etTransfo";

declare namespace tei="http://www.tei-c.org/ns/1.0";
declare namespace xi="http://www.w3.org/2001/XInclude";

import module namespace abbrev-index="http://fontane-nb.dariah.eu/abbrev-index" at "abbrev-index.xqm";
import module namespace config="http://textgrid.de/ns/SADE/config" at "../../config/config.xqm";
import module namespace fontaneSimple="http://fontane-nb.dariah.eu/teisimple" at "tei2teisimple.xqm";
import module namespace fsort="http://fontane-nb.dariah.eu/sort" at "sort.xqm";
import module namespace functx = "http://www.functx.com";
import module namespace macro-sort="http://fontane-nb.dariah.eu/macro-sort" at "macro-sort.xqm";
import module namespace prepCom="http://fontane-nb.dariah.eu/prepCom" at "prepcom.xqm";
import module namespace presort="http://fontane-nb.dariah.eu/presort" at "presort.xqm";
import module namespace simple2xhtml="http://fontane-nb.dariah.eu/simple2xhtml" at "simple2xhtml.xqm";
import module namespace tidySimple ="http://fontane-nb.dariah.eu/tidysimple" at "tidysimple.xqm";

declare variable $etTransfo:cases :=
    (
(:    "3qtcz.xml", (: case C :):)
    "3qtqv.xml" (: case A :)
(:    "3qtqw.xml" (: case B :) :)
(:    "3qtqx.xml" (: case D :):)
(:    "3qtqz.xml" (: case E :):)
    );

declare variable $etTransfo:dir := "/db/apps/SADE/modules/fontane/edited-text/";


(:~
 : The main function
 :
 : @author Michelle Weidling
 : @return An XHTML page for each notebook in /db/sade-projects/textgrid/data/print/xhtml
 : @return An intermediate TEI/XML with all notebooks in it in /db/sade-projects/textgrid/data/print/xml/fontane-full.xml
 :)
declare function etTransfo:complete() {
    for $case in $etTransfo:cases return
        (etTransfo:create-case($case),
        etTransfo:create-htmls($case)),

    try {
        etTransfo:create-print-tei()
    } catch * {
        error(QName("error", "ETTRANSFO02"), "An error occured while creating the whole TEI base for print.")
    },
    etTransfo:tidy-logs()
};


(:~
 : Resolves all XIncludes in a given case file and copies all notebooks belonging to a case into one file.
 :
 : @author Michelle Weidling
 : @param $showcase The filename of a showcase, e.g. "12345.xml"
 : @return The full path to the created showcase with all notebooks in it.
 :)
declare function etTransfo:create-case($showcase as xs:string) as xs:string {
    let $doc :=
        try {
            doc("/db/sade-projects/textgrid/data/xml/data/" || $showcase)
        } catch * {
            error(QName("error", "ETTRANSFO03"), "An error occured while opening " || $showcase || ".")
        }
    let $new-filename := $doc//tei:title[1] => replace(" ", "-") || ".xml"
    let $assure-dir-available :=
        if(xmldb:collection-available($config:data-root || "/print/")) then
            ()
        else
            (xmldb:create-collection($config:data-root, "/print"),
            xmldb:create-collection($config:data-root || "/print/", "xml"),
            xmldb:create-collection($config:data-root || "/print/", "xhtml"))

    let $create-new-file := xmldb:store($config:data-root || "/print/xml/", $new-filename, $doc/*)

    let $xis := doc($create-new-file)//xi:include
    let $update :=
        for $xi in $xis
            let $uri := $xi/@href/string()
            let $tei-referred-to :=
            try {
                doc("/db/sade-projects/textgrid/data/xml/data/" || $uri || ".xml")/tei:TEI
            } catch * {
                error(QName("error", "ETTRANSFO04"), "Couldn't find the node tei:TEI of  " || $uri || ".")
            }
            let $updated-notebook :=
                element tei:TEI {
                  attribute id {$uri},
                  $tei-referred-to/@*,
                  $tei-referred-to/*
              }

            return update replace $xi with $updated-notebook
    return
        $create-new-file
};


(:~
 : Gets all notebooks of a showcase in-memory.
 :
 : @author Michelle Weidling
 : @param $showcase The filename of a showcase, e.g. "12345.xml"
 : @return All notebooks of a showcase in-memory, i.e. a sequence of tei:TEI
 :)
declare function etTransfo:get-teis($showcase as xs:string) as element(tei:TEI)+ {
    let $doc := doc("/db/sade-projects/textgrid/data/xml/data/" || $showcase)
    let $summary-file-name := $doc//tei:title[1] => replace(" ", "-") || ".xml"
    let $summary-file :=
        try {
            doc($config:data-root || "/print/xml/" || $summary-file-name)
        } catch * {
            error(QName("error", "ETTRANSFO04"), "Summary file for case " || $showcase || "couldn't be opened.")
        }

    return
        $summary-file//tei:TEI
};


(:~
 : Creates the intermediate format of a given case and transforms each notebook of a showcase to a XHTML page.
 :
 : @author Michelle Weidling
 : @param $showcase The filename of a showcase, e.g. "12345.xml"
 : @return One or more strings indicating the location where the XHTML(s) have been stored to.
 :)
declare function etTransfo:create-htmls($showcase as xs:string) as xs:string+ {
    let $teis := etTransfo:get-teis($showcase)

    for $tei in $teis return
        let $log := etTransfo:create-log(substring-after($tei//tei:idno[@type = "TextGrid"], "textgrid:"))
        return
            etTransfo:transform-tei($tei, $log)
};


(:~
 : Handles the transformation to the intermediate format and the XHTML
 : representation.
 :
 : @param $tei The current notebook as element(tei:TEI)
 :)
declare function etTransfo:transform-tei($tei as element(tei:TEI), $log as xs:string) {
(:    macro-sort:main($tei, $tei/@id):)
(:    => fsort:main():)
    fsort:main($tei, $log)
    => prepCom:main($tei/@id)
    => fontaneSimple:main($tei/@id, $log)
    => tidySimple:main($tei/@id)
    => simple2xhtml:main($tei/@id)
};


(:~
 : Assembles the intermediate format version of each notebook and copies them into one file called "fontane-full.xml".
 : Furthermore, additional data needed for the print is appended.
 :
 : @author Michelle Weidling
 : @return A string indicating the location where the full TEI/XML of the edition was stored to.
 :)
declare function etTransfo:create-print-tei() as xs:string {
    let $xmls := etTransfo:get-all-xmls()
    let $complete-print-file :=
        element {QName("http://www.tei-c.org/ns/1.0", "teiCorpus")} {
            $xmls,
            try {
                etTransfo:get-literature()
            } catch * {
                error(QName("error", "ETTRANSFO06"), "An error occured while creating the bibliography.")
            },

            try {
                abbrev-index:main()
            } catch * {
                error(QName("error", "ETTRANSFO07"), "An error occured while creating the index of abbreviations.")
            }
        }
    return
        xmldb:store($config:data-root || "/print/xml/", "fontane-full.xml", $complete-print-file)
};

(:~
 : An auxiliary function. Returns all final intermediate format TEI/XML files.
 :
 : @author Michelle Weidling
 : @return All available final intermediate format TEI/XML files as opened documents
 :)
declare function etTransfo:get-all-xmls() as node()+ {
    for $res in collection($config:data-root || "/print/xml/")
        [contains(base-uri(.), ".xml")
        and not(contains(substring-after(base-uri(.), "/print/xml/"), "-"))]
    order by $res/tei:TEI/@key1, number($res/tei:TEI/@key2)
    return
       $res
};


(:~
 : Creates a summary tei:div with all information needed for the print
 : bibliography.
 :
 : @author Michelle Weidling
 : @return a tei:div containing all bibliographic sections of the print
 :)
declare function etTransfo:get-literature() as element(tei:div) {
    element {QName("http://www.tei-c.org/ns/1.0", "text")} {
        attribute type {"bibliography"},

        element tei:div {
            attribute n {"Fontanes_Werke"},
            local:make-bib-div("Erstausgaben"),
            local:make-bib-div("Werkausgaben"),
            local:make-bib-div("Briefausgaben"),
            local:make-bib-div("Postume_Einzelausgaben"),
            local:make-bib-div("Populaere_postume_Einzelausgaben")
        },
        local:make-bib-div("Fontanes_Quellen"),
        local:make-bib-div("Allgemeine_Nachschlagewerke"),
        local:make-bib-div("Nachschlagewerke_der_Fontane-Forschung"),
        local:make-bib-div("Forschungsliteratur")
    }
};


(:~
 : An auxiliary function to create the tei:div elements in the bibliography.
 :
 : @author Michelle Weidling
 : @param $identifier An identifier for a subtype of scholarly literature
 : @return a tei:div containing all bibliographic information of the literatur's subtype
 :)
declare function local:make-bib-div($identifier as xs:string) as element(tei:div) {
    element tei:div {
        attribute n {$identifier},
        etTransfo:make-bib-entries($identifier)
    }
};



(:~
 : Creates a div containing all the scholarly literature ("Forschungsliteratur")
 : of a kind that's listed in the respective index. Examples for this are
 : "Allgemeine Nachschlagewerke" and "Werkausgaben"
 :
 : @author Michelle Weidling
 : @param $type The type of a tei:listBibl, e.g. "Forschungsliteratur"
 : @return a tei:div containing the scribal abbreviation as well as the full
 : bibliographic reference
 :)
declare function etTransfo:make-bib-entries($type as xs:string) as element(tei:div)* {
    for $entry in doc("/db/sade-projects/textgrid/data/xml/data/25547.xml")//tei:listBibl[@type = $type]//tei:bibl
        let $abbr :=
            if($entry/tei:abbr) then
                $entry/tei:abbr
            else
                $entry/tei:choice/tei:abbr
        let $full-ref := $entry/tei:choice/tei:expan
        let $bib-no := count($entry/ancestor::tei:bibl)

        return
            element tei:div {
                attribute type {"entry"},
                element tei:seg {
                    attribute type {"abbr"},
                    $abbr
                },
                if($full-ref) then
                    element tei:seg {
                        attribute type {"full"},
                        $full-ref
                    }

                else if($entry/tei:ptr[@target]) then
                    let $target-id := replace($entry/tei:ptr/@target, "#", "")
                    let $referenced-entry := $entry/ancestor::tei:body//tei:bibl[@xml:id = $target-id]
                    let $same-as := $referenced-entry/tei:choice/tei:abbr/string()
                    let $same-as-loc := $referenced-entry/ancestor::tei:listBibl[1]/@type/string()

                    return
                        (element tei:seg {
                            attribute type {"same-as"},
                            $same-as
                        },
                        element tei:seg {
                            attribute type {"same-as-loc"},
                            $same-as-loc
                        })

                else
                    (),
                if($bib-no gt 0) then
                    element tei:seg {
                        attribute type {"sub-entry"},
                        $bib-no
                    }
                else
                    ()
            }
};


declare function etTransfo:create-log($uri as xs:string) as xs:string {
  let $assure-dir-available := etTransfo:assure-dir-available("logs")
  let $log-name := $uri || "-log.xml"
  return xmldb:store($etTransfo:dir || "logs", $log-name, <log/>)
};


declare function etTransfo:assure-dir-available($dir-name as xs:string) {
    if(xmldb:collection-available($etTransfo:dir || "logs")) then
        $etTransfo:dir || "logs"
    else
        xmldb:create-collection($etTransfo:dir, "logs")

};


(:~
 : Only the logs which acutally contain info should be kept.
 :
 :)
declare function etTransfo:tidy-logs() {
    for $log in collection($etTransfo:dir || "logs/") return
        if($log//LogEntry) then
            ()
        else
            let $uri := substring-after(base-uri($log), "logs/")
            return
                xmldb:remove($etTransfo:dir || "logs/", $uri)
};
